@@ -1,17 +1,5 @@
-/*=========================================================================
-
-  Program:   Visualization Toolkit
-  Module:    vtkSetGet.h
-
-  Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
-  All rights reserved.
-  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.
-
-     This software is distributed WITHOUT ANY WARRANTY; without even
-     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-     PURPOSE.  See the above copyright notice for more information.
-
-=========================================================================*/
+// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
+// SPDX-License-Identifier: BSD-3-Clause
 /**
  * @class   SetGet
  *
@@ -204,6 +192,13 @@
 #define vtkSetStringMacroOverride(name)                                                            \
   void Set##name(const char* _arg) vtkSetStringBodyMacro(name, _arg) override
 
+//
+// Set a string token. Creates member Set"name"()
+// (e.g., SetArrayName(vtkStringToken));
+//
+#define vtkSetStringTokenMacro(name)                                                               \
+  virtual void Set##name(vtkStringToken _arg) vtkSetStringTokenBodyMacro(name, _arg)
+
 // Set a file path, like vtkSetStringMacro but with VTK_FILEPATH hint.
 #define vtkSetFilePathMacro(name)                                                                  \
   virtual void Set##name(VTK_FILEPATH const char* _arg) vtkSetStringBodyMacro(name, _arg)
@@ -248,6 +243,26 @@
 //
 #define vtkGetStringMacro(name) virtual char* Get##name() vtkGetStringBodyMacro(name)
 
+//
+// Get string token.  Creates member Get"name"()
+// (e.g., vtkStringToken GetArrayName());
+//
+#define vtkGetStringTokenMacro(name)                                                               \
+  virtual vtkStringToken Get##name() vtkGetStringTokenBodyMacro(name)
+
+// This macro defines a body of set string macro. It can be used either in
+// the header file using vtkSetStringMacro or in the implementation.
+#define vtkSetStringTokenBodyMacro(name, _arg)                                                     \
+  {                                                                                                \
+    vtkDebugMacro(<< " setting " #name " to " << _arg.Data());                                     \
+    if (this->name == _arg)                                                                        \
+    {                                                                                              \
+      return;                                                                                      \
+    }                                                                                              \
+    this->name = _arg;                                                                             \
+    this->Modified();                                                                              \
+  }
+
 // Get a file path, like vtkGetStringMacro but with VTK_FILEPATH hint.
 #define vtkGetFilePathMacro(name)                                                                  \
   virtual VTK_FILEPATH VTK_FUTURE_CONST char* Get##name()                                          \
@@ -261,6 +276,14 @@
     return this->name;                                                                             \
   }
 
+// This macro defines a body of get string-token macro. It can be used either in
+// the header file using vtkGetStringTokenMacro or in the implementation.
+#define vtkGetStringTokenBodyMacro(name)                                                           \
+  {                                                                                                \
+    vtkDebugMacro(<< " returning " #name " of " << this->name.Data());                             \
+    return this->name;                                                                             \
+  }
+
 //
 // Set std::string. Creates a member Set"name"()
 //
@@ -270,8 +293,16 @@
     vtkDebugMacro(<< " setting " #name " to " << (arg ? arg : "(null)"));                          \
     if (arg)                                                                                       \
     {                                                                                              \
+      if (this->name == arg)                                                                       \
+      {                                                                                            \
+        return;                                                                                    \
+      }                                                                                            \
       this->name = arg;                                                                            \
     }                                                                                              \
+    else if (this->name.empty())                                                                   \
+    {                                                                                              \
+      return;                                                                                      \
+    }                                                                                              \
     else                                                                                           \
     {                                                                                              \
       this->name.clear();                                                                          \
@@ -284,8 +315,16 @@
     vtkDebugMacro(<< " setting " #name " to " << (arg ? arg : "(null)"));                          \
     if (arg)                                                                                       \
     {                                                                                              \
+      if (this->name == arg)                                                                       \
+      {                                                                                            \
+        return;                                                                                    \
+      }                                                                                            \
       this->name = arg;                                                                            \
     }                                                                                              \
+    else if (this->name.empty())                                                                   \
+    {                                                                                              \
+      return;                                                                                      \
+    }                                                                                              \
     else                                                                                           \
     {                                                                                              \
       this->name.clear();                                                                          \
@@ -742,6 +781,7 @@
 // This is to avoid vtkObject #include of vtkOutputWindow
 // while vtkOutputWindow #includes vtkObject
 
+VTK_ABI_NAMESPACE_BEGIN
 extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayText(const char*);
 extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayErrorText(const char*);
 extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayWarningText(const char*);
@@ -759,6 +799,7 @@ extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayGenericWarningText(
   const char*, int, const char*);
 extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayDebugText(
   const char*, int, const char*, vtkObject* sourceObj);
+VTK_ABI_NAMESPACE_END
 
 //
 // This macro is used for any output that may not be in an instance method
@@ -769,11 +810,11 @@ extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayDebugText(
   {                                                                                                \
     if (vtkObject::GetGlobalWarningDisplay())                                                      \
     {                                                                                              \
-      vtkOStreamWrapper::EndlType endl;                                                            \
+      vtkOStreamWrapper::EndlType const endl;                                                      \
       vtkOStreamWrapper::UseEndl(endl);                                                            \
       vtkOStrStreamWrapper vtkmsg;                                                                 \
       vtkmsg << "" x;                                                                              \
-      std::string _filename = vtksys::SystemTools::GetFilenameName(__FILE__);                      \
+      std::string const _filename = vtksys::SystemTools::GetFilenameName(__FILE__);                \
       vtkOutputWindowDisplayGenericWarningText(_filename.c_str(), __LINE__, vtkmsg.str());         \
       vtkmsg.rdbuf()->freeze(0);                                                                   \
     }                                                                                              \
@@ -809,7 +850,7 @@ extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayDebugText(
   {                                                                                                \
     if (vtkObject::GetGlobalWarningDisplay())                                                      \
     {                                                                                              \
-      vtkOStreamWrapper::EndlType endl;                                                            \
+      vtkOStreamWrapper::EndlType const endl;                                                      \
       vtkOStreamWrapper::UseEndl(endl);                                                            \
       vtkOStrStreamWrapper vtkmsg;                                                                 \
       vtkObject* _object = const_cast<vtkObject*>(static_cast<const vtkObject*>(self));            \
@@ -822,7 +863,7 @@ extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayDebugText(
         vtkmsg << "(nullptr): ";                                                                   \
       }                                                                                            \
       vtkmsg << "" x;                                                                              \
-      std::string _filename = vtksys::SystemTools::GetFilenameName(__FILE__);                      \
+      std::string const _filename = vtksys::SystemTools::GetFilenameName(__FILE__);                \
       vtkOutputWindowDisplayErrorText(_filename.c_str(), __LINE__, vtkmsg.str(), _object);         \
       vtkmsg.rdbuf()->freeze(0);                                                                   \
       vtkObject::BreakOnError();                                                                   \
@@ -841,7 +882,7 @@ extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayDebugText(
   {                                                                                                \
     if (vtkObject::GetGlobalWarningDisplay())                                                      \
     {                                                                                              \
-      vtkOStreamWrapper::EndlType endl;                                                            \
+      vtkOStreamWrapper::EndlType const endl;                                                      \
       vtkOStreamWrapper::UseEndl(endl);                                                            \
       vtkOStrStreamWrapper vtkmsg;                                                                 \
       vtkObject* _object = const_cast<vtkObject*>(static_cast<const vtkObject*>(self));            \
@@ -854,7 +895,7 @@ extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayDebugText(
         vtkmsg << "(nullptr): ";                                                                   \
       }                                                                                            \
       vtkmsg << "" x;                                                                              \
-      std::string _filename = vtksys::SystemTools::GetFilenameName(__FILE__);                      \
+      std::string const _filename = vtksys::SystemTools::GetFilenameName(__FILE__);                \
       vtkOutputWindowDisplayWarningText(_filename.c_str(), __LINE__, vtkmsg.str(), _object);       \
       vtkmsg.rdbuf()->freeze(0);                                                                   \
     }                                                                                              \
@@ -879,7 +920,7 @@ extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayDebugText(
     vtkObject* _object = const_cast<vtkObject*>(static_cast<const vtkObject*>(self));              \
     if ((!_object || _object->GetDebug()) && vtkObject::GetGlobalWarningDisplay())                 \
     {                                                                                              \
-      vtkOStreamWrapper::EndlType endl;                                                            \
+      vtkOStreamWrapper::EndlType const endl;                                                      \
       vtkOStreamWrapper::UseEndl(endl);                                                            \
       vtkOStrStreamWrapper vtkmsg;                                                                 \
       if (_object)                                                                                 \
@@ -891,7 +932,7 @@ extern VTKCOMMONCORE_EXPORT void vtkOutputWindowDisplayDebugText(
         vtkmsg << "(nullptr): ";                                                                   \
       }                                                                                            \
       vtkmsg << "" x;                                                                              \
-      std::string _filename = vtksys::SystemTools::GetFilenameName(__FILE__);                      \
+      std::string const _filename = vtksys::SystemTools::GetFilenameName(__FILE__);                \
       vtkOutputWindowDisplayDebugText(_filename.c_str(), __LINE__, vtkmsg.str(), _object);         \
       vtkmsg.rdbuf()->freeze(0);                                                                   \
     }                                                                                              \
@@ -1233,6 +1274,22 @@ public:
 #define VTK_FALLTHROUGH ((void)0)
 #endif
 
+//----------------------------------------------------------------------------
+// To suppress code with undefined behaviour. Ideally, such code should be fixed
+// but sometimes suppression can be useful.
+
+#if defined(__has_attribute)
+#if __has_attribute(no_sanitize)
+#define VTK_NO_UBSAN __attribute__((no_sanitize("undefined")))
+#else
+#define VTK_NO_UBSAN
+#endif
+#else
+#define VTK_NO_UBSAN
+#endif
+
+//----------------------------------------------------------------------------
+
 // XXX(xcode-8)
 // AppleClang first supported thread_local only by Xcode 8, for older Xcodes
 // fall back to the non-standard __thread which is equivalent for many, but
